@@ -562,15 +562,40 @@ get_sorted_list_by_action(Plisthead *pkgs, action_t action)
 
 #define H_BUF 6
 
+/*
+ * Return list of installed core packages that should be checked for upgrade
+ * first, prior to any general upgrades.
+ */
+static char **
+get_core_pkgs(void)
+{
+	const char *pkgs[] = {
+		"pkg_install",
+		"pkgin",
+		NULL,
+	};
+	size_t n = 0, p;
+	char **corepkgs = NULL;
+
+	for (p = 0; pkgs[p] != NULL; p++) {
+		if (find_local_pkg(pkgs[p], NULL) != NULL) {
+			corepkgs = xrealloc(corepkgs, (n + 2) * sizeof(char *));
+			corepkgs[n] = xstrdup(pkgs[p]);
+			corepkgs[++n] = NULL;
+		}
+	}
+
+	return corepkgs;
+}
+
 int
 pkgin_install(char **pkgargs, int do_inst, int upgrade)
 {
-	Plistarray	*conflicts;
 	FILE		*fp;
 	int		installnum = 0, upgradenum = 0;
 	int		refreshnum = 0, downloadnum = 0;
 	int		removenum = 0, supersedenum = 0;
-	int		conflictnum = 0;
+	int		coreupg = 0;
 	int		argn, rc = EXIT_SUCCESS;
 	int		privsreqd = PRIVS_PKGINDB;
 	uint64_t	free_space;
@@ -578,12 +603,14 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 	size_t		len;
 	ssize_t		llen;
 	Pkglist		*p;
-	Plisthead	*impacthead, *downloadhead = NULL, *installhead = NULL;
+	Plisthead	*impacthead = NULL;
+	Plisthead	*downloadhead = NULL, *installhead = NULL;
 	char		**names;
 	char		*toinstall = NULL, *toupgrade = NULL;
 	char		*torefresh = NULL, *todownload = NULL;
 	char		*toremove = NULL, *tosupersede = NULL;
 	char		*unmet_reqs = NULL;
+	char		**corepkgs;
 	char		pkgrepo[BUFSIZ], query[BUFSIZ];
 	char		h_psize[H_BUF], h_fsize[H_BUF], h_free[H_BUF];
 	struct		stat st;
@@ -600,12 +627,32 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 		errx(EXIT_FAILURE, MSG_DONT_HAVE_RIGHTS);
 
 	/*
-	 * Calculate full impact (list of packages and actions to perform) of
-	 * requested operation.
+	 * Initialise local CONFLICTS as it will be used by pkg_impact().
 	 */
-	if ((impacthead = pkg_impact(pkgargs, &rc)) == NULL) {
-		printf(MSG_NOTHING_TO_DO);
-		return rc;
+	l_conflicthead = init_array(CONFLICTS_HASH_SIZE);
+	pkgindb_doquery(LOCAL_CONFLICTS, record_pattern_to_array, l_conflicthead);
+
+	/*
+	 * If pkgargs is NULL we're performing an upgrade.  First check to see
+	 * if there are any upgrades for core package tools, and if so perform
+	 * them first.
+	 */
+	if (pkgargs == NULL) {
+		if ((corepkgs = get_core_pkgs()) != NULL) {
+			impacthead = pkg_impact(corepkgs, &rc, 0);
+			if (impacthead != NULL) {
+				coreupg = 1;
+			}
+		}
+	}
+	/*
+	 * Otherwise calculate full impact of either full upgrade or install.
+	 */
+	if (impacthead == NULL) {
+		if ((impacthead = pkg_impact(pkgargs, &rc, 1)) == NULL) {
+			printf(MSG_NOTHING_TO_DO);
+			return rc;
+		}
 	}
 
 	/* check for required files */
@@ -615,25 +662,6 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 				unmet_reqs =
 				    action_list(unmet_reqs, p->rpkg->full);
 
-	/*
-	 * Check for conflicts.
-	 *
-	 * XXX: this should really be done during impact so that we have a
-	 * chance to consider alternatives.
-	 */
-	conflicts = init_array(CONFLICTS_HASH_SIZE);
-	get_conflicts(conflicts);
-	SLIST_FOREACH(p, impacthead, next) {
-		if (!action_is_install(p->action))
-			continue;
-		if (pkg_has_conflicts(p, conflicts))
-			conflictnum++;
-	}
-	free_array(conflicts);
-
-	if (conflictnum && !check_yesno(DEFAULT_NO))
-		goto installend;
-
 	/*
 	 * Set up counters.
 	 */
@@ -917,6 +945,13 @@ pkgin_install(char **pkgargs, int do_inst, int upgrade)
 
 	(void)update_db(LOCAL_SUMMARY, 1);
 
+	/*
+	 * If we only upgraded the package tools and it was successful then
+	 * print a message about performing the subsequent full upgrade.
+	 */
+	if (coreupg && rc == EXIT_SUCCESS)
+		printf(MSG_PKGTOOLS_UPGRADED);
+
 installend:
 	XFREE(todownload);
 	XFREE(toinstall);
