@@ -159,7 +159,7 @@ sum_close(struct archive *a, void *data)
  * Download a package to the local cache.
  */
 off_t
-download_pkg(char *pkg_url, FILE *fp)
+download_pkg(char *pkg_url, FILE *fp, int cur, int total)
 {
 	struct url_stat st;
 	size_t size, wrote;
@@ -168,7 +168,7 @@ download_pkg(char *pkg_url, FILE *fp)
 	struct url *url;
 	fetchIO *f = NULL;
 	char buf[4096];
-	char *pkg, *ptr;
+	char *pkg, *ptr, *msg;
 
 	if ((url = fetchParseURL(pkg_url)) == NULL)
 		errx(EXIT_FAILURE, "%s: parse failure", pkg_url);
@@ -176,8 +176,10 @@ download_pkg(char *pkg_url, FILE *fp)
 	if ((f = fetchXGet(url, &st, fetchflags)) == NULL) {
 		fprintf(stderr, "download error: %s %s\n", pkg_url,
 		    fetchLastErrString);
+		fetchFreeURL(url);
 		return -1;
 	}
+	fetchFreeURL(url);
 
 	if ((pkg = strrchr(pkg_url, '/')) != NULL)
 		pkg++;
@@ -185,11 +187,11 @@ download_pkg(char *pkg_url, FILE *fp)
 		pkg = (char *)pkg_url; /* should not happen */
 
 	if (parsable) {
-		printf("downloading %s", pkg);
+		printf("[%d/%d] downloading %s", cur, total, pkg);
 	} else {
-		printf("downloading %s:   0%%", pkg);
+		msg = xasprintf("[%d/%d] %s", cur, total, pkg);
 		fflush(stdout);
-		start_progress_meter(pkg, st.size, &statsize);
+		start_progress_meter(msg, st.size, &statsize);
 	}
 
 	while (written < st.size) {
@@ -219,11 +221,12 @@ download_pkg(char *pkg_url, FILE *fp)
 
 	if (parsable)
 		printf(" done.\n");
-	else
+	else {
 		stop_progress_meter();
+		free(msg);
+	}
 
 	fetchIO_close(f);
-	fetchFreeURL(url);
 
 	if (written != st.size) {
 		fprintf(stderr, "download error: %s truncated\n", pkg_url);
