@@ -118,6 +118,7 @@ calculate_action(Pkglist *lpkg, Pkglist *rpkg)
  *
  * argv0: SUPERSEDES pattern
  * argv1: PKGBASE, may be NULL if it cannot be determined from pattern
+ * argv2: PKGNAME of replacement
  */
 static int
 record_supersedes(void *param, int argc, char **argv, char **colname)
@@ -145,7 +146,7 @@ record_supersedes(void *param, int argc, char **argv, char **colname)
 		return PDB_OK;
 
 	/*
-	 * Look to see if we already found this package via an different match.
+	 * Look to see if we already found this package via a different match.
 	 */
 	SLIST_FOREACH(p, supersedes, next) {
 		if (strcmp(lpkg->full, p->lpkg->full) == 0)
@@ -161,6 +162,7 @@ record_supersedes(void *param, int argc, char **argv, char **colname)
 	p->patterns[1] = NULL;
 	p->patcount = 1;
 	p->lpkg = lpkg;
+	p->replace = xstrdup(argv[2]);
 	SLIST_INSERT_HEAD(supersedes, p, next);
 
 	return PDB_OK;
@@ -176,19 +178,8 @@ find_supersedes(Plistarray *impacthead)
 
 	supersedes = init_head();
 
-	/*
-	 * Only consider packages that are going to be installed.
-	 */
-	for (i = 0; i < impacthead->size; i++) {
-	SLIST_FOREACH(pkg, &impacthead->head[i], next) {
-		if (!action_is_install(pkg->action))
-			continue;
-
-		sqlite3_snprintf(BUFSIZ, query, REMOTE_SUPERSEDES,
-		    pkg->rpkg->full);
-		pkgindb_doquery(query, record_supersedes, supersedes);
-	}
-	}
+	sqlite3_snprintf(BUFSIZ, query, REMOTE_SUPERSEDES);
+	pkgindb_doquery(query, record_supersedes, supersedes);
 
 	if (SLIST_EMPTY(supersedes)) {
 		free_pkglist(&supersedes);
@@ -205,8 +196,9 @@ find_supersedes(Plistarray *impacthead)
 static action_t
 add_remote_to_impact(Plistarray *impacthead, Pkglist *pkg)
 {
-	Pkglist *lpkg;
+	Pkglist *lpkg, *p;
 	size_t slot;
+	char *cmatch;
 
 	pkg->action = ACTION_NONE;
 
@@ -224,6 +216,34 @@ add_remote_to_impact(Plistarray *impacthead, Pkglist *pkg)
 		pkg->action = calculate_action(pkg->lpkg, pkg->rpkg);
 	}
 
+	/*
+	 * If incoming package matches a local CONFLICTS entry then find the
+	 * corresponding local package and mark it for removal.
+	 */
+	if ((cmatch = pkg_conflicts(pkg)) != NULL) {
+		/*
+		 * Look up corresponding local package for this match.
+		 *
+		 * TODO: This should be optimised to be already stored in
+		 * l_conflicthead and avoid the need for additional queries.
+		 */
+		char query[BUFSIZ];
+		char *cpkgname = xmalloc(BUFSIZ * sizeof(char));
+		sqlite3_snprintf(BUFSIZ, query, REMOTE_CONFLICTS, cmatch);
+		pkgindb_doquery(query, pdb_get_value, cpkgname);
+		if ((lpkg = find_local_pkg(cpkgname, NULL)) != NULL) {
+			if (p = local_pkg_in_impact(impacthead, lpkg)) {
+				p->action = ACTION_REMOVE;
+			} else {
+				p = malloc_pkglist();
+				p->action = ACTION_REMOVE;
+				p->lpkg = lpkg;
+				slot = pkg_hash_entry(lpkg->name, impacthead->size);
+				SLIST_INSERT_HEAD(&impacthead->head[slot], p, next);
+			}
+		}
+	}
+
 	slot = pkg_hash_entry(pkg->rpkg->name, impacthead->size);
 	SLIST_INSERT_HEAD(&impacthead->head[slot], pkg, next);
 
@@ -273,7 +293,7 @@ finish_deps_spinner(int istty)
  * through all local packages and check each for upgrades.
  */
 static Plistarray *
-pkg_impact_upgrade(void)
+pkg_impact_upgrade(int verbose)
 {
 	Plistarray *deps, *impacthead;
 	Plisthead *supersedes;
@@ -323,11 +343,46 @@ pkg_impact_upgrade(void)
 		 * remote package dependencies as there may be new or updated
 		 * dependencies that need to be installed.
 		 */
-		update_deps_spinner(istty);
+		if (verbose)
+			update_deps_spinner(istty);
 		get_depends_recursive(p->rpkg->full, deps, DEPENDS_REMOTE);
 	}
 	}
 
+	/*
+	 * Get SUPERSEDES entries matching local packages.  For each affected
+	 * package, mark as superseded which will cause it to be removed, and
+	 * add replacement if not already installed.
+	 *
+	 */
+	if ((supersedes = find_supersedes(impacthead))) {
+		SLIST_FOREACH(lpkg, supersedes, next) {
+			Pkglist *oldp, *newp;
+			/*
+			 * find_supersedes() ensures lpkg will be set and this
+			 * will return a valid entry.
+			 */
+			oldp = local_pkg_in_impact(impacthead, lpkg->lpkg);
+			oldp->action = ACTION_SUPERSEDED;
+
+			/*
+			 * If the replacement is not already installed then install it.
+			 */
+			char *repmatch = xasprintf("%s-[0-9]*", lpkg->replace);
+			if ((newp = find_local_pkg(repmatch, lpkg->replace)) == NULL) {
+				newp = malloc_pkglist();
+				newp->rpkg = find_remote_pkg(repmatch, lpkg->replace, NULL);
+				newp->keep = oldp->lpkg->keep;
+				if (verbose)
+					update_deps_spinner(istty);
+				get_depends_recursive(newp->rpkg->full, deps, DEPENDS_REMOTE);
+				add_remote_to_impact(impacthead, newp);
+			}
+		}
+		free_pkglist(&supersedes);
+	}
+
+
 	/*
 	 * We now have a full list of dependencies for all local packages,
 	 * process them in turn, adding to impact list.  As we may have already
@@ -348,22 +403,6 @@ pkg_impact_upgrade(void)
 	}
 	free_array(deps);
 
-	/*
-	 * Get SUPERSEDES entries matching local packages.  For each affected
-	 * package, mark as superseded which will cause it to be removed.
-	 */
-	if ((supersedes = find_supersedes(impacthead))) {
-		SLIST_FOREACH(lpkg, supersedes, next) {
-			/*
-			 * find_supersedes() ensures lpkg will be set and this
-			 * will return a valid entry.
-			 */
-			p = local_pkg_in_impact(impacthead, lpkg->lpkg);
-			p->action = ACTION_SUPERSEDED;
-		}
-		free_pkglist(&supersedes);
-	}
-
 	return impacthead;
 }
 
@@ -477,7 +516,7 @@ recurse_upgrades(Plisthead *upgrades, Plistarray *impacthead)
  * packages and their dependencies to install.
  */
 static Plistarray *
-pkg_impact_install(char **pkgargs, int *rc)
+pkg_impact_install(char **pkgargs, int *rc, int verbose)
 {
 	Plistarray *deps, *impacthead;
 	Plisthead *ipkgs;
@@ -532,7 +571,8 @@ pkg_impact_install(char **pkgargs, int *rc)
 		 * calculate their actions based on whether they match a local
 		 * package or not.
 		 */
-		update_deps_spinner(istty);
+		if (verbose)
+			update_deps_spinner(istty);
 		deps = init_array(DEPS_HASH_SIZE);
 		get_depends_recursive(p->rpkg->full, deps, DEPENDS_REMOTE);
 		for (i = 0; i < deps->size; i++) {
@@ -589,7 +629,7 @@ pkg_impact_install(char **pkgargs, int *rc)
  * which loops through all local packages looking for upgrades.
  */
 Plisthead *
-pkg_impact(char **pkgargs, int *rc)
+pkg_impact(char **pkgargs, int *rc, int verbose)
 {
 	Plistarray *pkgs;
 	Plisthead *impacthead;
@@ -599,15 +639,17 @@ pkg_impact(char **pkgargs, int *rc)
 	istty = isatty(fileno(stdout));
 
 	TRACE("[>]-entering impact\n");
-	start_deps_spinner(istty);
+	if (verbose)
+		start_deps_spinner(istty);
 
 	if (pkgargs)
-		pkgs = pkg_impact_install(pkgargs, rc);
+		pkgs = pkg_impact_install(pkgargs, rc, verbose);
 	else
-		pkgs = pkg_impact_upgrade();
+		pkgs = pkg_impact_upgrade(verbose);
 
 	TRACE("[<]-leaving impact\n");
-	finish_deps_spinner(istty);
+	if (verbose)
+		finish_deps_spinner(istty);
 
 	impacthead = init_head();
 
