@@ -39,10 +39,12 @@ int
 find_preferred_pkg(const char *pkgname, Pkglist **pkg, char **match)
 {
 	Pkglist *p, *best = NULL;
+	int i;
 	char *result = NULL;
 
 	/* Find best match */
-	SLIST_FOREACH(p, &r_plisthead, next) {
+	for (i = 0; i < REMOTE_PKG_HASH_SIZE; i++) {
+	SLIST_FOREACH(p, &r_plisthead[i], next) {
 		if (!pkg_match(pkgname, p->full))
 			continue;
 
@@ -66,6 +68,7 @@ find_preferred_pkg(const char *pkgname, Pkglist **pkg, char **match)
 		if (best == NULL || version_check(best->full, p->full) == 2)
 			best = p;
 	}
+	}
 
 	/*
 	 * Save match if requested.  If there was a successful match then
@@ -122,32 +125,94 @@ unique_pkg(const char *pkgname, const char *dest)
 }
 
 /*
- * Return best matching SLIST entry in package list, or NULL if no match.
+ * Return best remote package for a pattern and optional pkgpath, or NULL if no
+ * valid match.
  */
 Pkglist *
-find_pkg_match(Plisthead *plisthead, char *match)
+find_remote_pkg(const char *pattern, const char *pkgname, const char *pkgpath)
 {
-	Pkglist	*pkg = NULL, *p;
+	Pkglist	*p, *pkg = NULL;
+	size_t size = REMOTE_PKG_HASH_SIZE;
+	size_t slot = 0;
 
-	SLIST_FOREACH(p, plisthead, next) {
-		if (!pkg_match(match, p->full))
-			continue;
+	/*
+	 * If a valid pkgname is supplied then the caller has indicated that
+	 * the pattern will always match the specific pkgname, and so we can
+	 * optimise the lookup by only considering the hash entry for that
+	 * pkgname.  For example pattern is "foo>=1<2" and pkgname is "foo".
+	 */
+	if (pkgname) {
+		slot = pkg_hash_entry(pkgname, REMOTE_PKG_HASH_SIZE);
+		size = slot + 1;
+	}
 
-		if (chk_preferred(p->full, NULL) != 0) {
-			printf("no chk_pref %s\n", p->full);
-			continue;
-		}
+	for (; slot < size; slot++) {
+		SLIST_FOREACH(p, &r_plisthead[slot], next) {
+			if (!pkg_match(pattern, p->full))
+				continue;
 
-		/* Save best match */
-		if (pkg == NULL || version_check(pkg->full, p->full) == 2) {
-			TRACE("Matched %s with %s\n", match, p->full);
-			pkg = p;
+			/*
+			 * If pkgpath is specified then the remote entry must
+			 * match, to avoid newer releases being pulled in when
+			 * the user may have specified a particular version in
+			 * the past.
+			 */
+			if (pkgpath && pkgstrcmp(pkgpath, p->pkgpath) != 0)
+				continue;
+
+			if (chk_preferred(p->full, NULL) != 0)
+				continue;
+
+			/*
+			 * Save match if we haven't seen one yet, or if this
+			 * one is a higher version number than the current best
+			 * match.
+			 */
+			if (pkg == NULL ||
+			    version_check(pkg->full, p->full) == 2)
+				pkg = p;
 		}
 	}
 
 	return pkg;
 }
 
+/*
+ * Find first match of a locally-installed package for a package pattern.
+ * Returns a Pkglist entry on success or NULL on failure.
+ *
+ * Just return the first match.  While technically we could find the "best"
+ * result for an alternate match it doesn't make any practical sense.
+ */
+Pkglist *
+find_local_pkg(const char *pattern, const char *pkgname)
+{
+	Pkglist	*p;
+	size_t size = LOCAL_PKG_HASH_SIZE;
+	size_t slot = 0;
+
+	/*
+	 * If a valid pkgname is supplied then the caller has indicated that
+	 * the pattern will always match the specific pkgname, and so we can
+	 * optimise the lookup by only considering the hash entry for that
+	 * pkgname.  For example pattern is "foo>=1<2" and pkgname is "foo".
+	 */
+	if (pkgname) {
+		slot = pkg_hash_entry(pkgname, LOCAL_PKG_HASH_SIZE);
+		size = slot + 1;
+	}
+
+	for (; slot < size; slot++) {
+		SLIST_FOREACH(p, &l_plisthead[slot], next) {
+			if (pkg_match(pattern, p->full)) {
+				return p;
+			}
+		}
+	}
+
+	return NULL;
+}
+
 /* basic full package format detection */
 int
 exact_pkgfmt(const char *pkgname)
@@ -165,26 +230,6 @@ exact_pkgfmt(const char *pkgname)
 	return isdigit((int)*p);
 }
 
-/* 
- * check whether or not pkgarg is a full pkgname (foo-1.0)
- */
-char *
-simple_pkg_match(Plisthead *plisthead, const char *pkgarg)
-{
-	Pkglist	*pkglist;
-
-	/* check for package existence */
-	SLIST_FOREACH(pkglist, plisthead, next) {
-		if (!pkg_match(pkgarg, pkglist->full))
-			continue;
-
-		return xstrdup(pkglist->full);
-	}
-
-
-	return NULL;
-}
-
 /* similar to opattern.c's pkg_order but without pattern */
 int
 version_check(char *first_pkg, char *second_pkg)
@@ -205,94 +250,6 @@ version_check(char *first_pkg, char *second_pkg)
 		return 2;
 }
 
-static void
-clear_pattern(char *depend)
-{
-	char *p;
-
-	if ((p = strpbrk(depend, GLOBCHARS)) != NULL)
-		*p = '\0';
-	else
-		return;
-
-	/* get rid of trailing dash if any */
-	if (*depend != '\0' && *--p == '-')
-		*p = '\0';
-}
-
-static void
-cleanup_version(char *pkgname)
-{
-	char	*exten;
-
-	/* have we got foo-something ? */
-	if ((exten = strrchr(pkgname, '-')) == NULL)
-		return;
-
-	/*
-	 * -something has a dot, it's a version number
-	 * unless it was something like clutter-gtk0.10
-	 */
-	if (isdigit((int)exten[1]) && strchr(exten, '.') != NULL)
-		*exten = '\0';
-}
-
-uint8_t
-non_trivial_glob(char *depend)
-{
-	if (charcount(depend, '[') > 1)
-		return 1;
-
-	return 0;
-}
-
-/*
- * AFAIK, here are the dewey/glob we can find as dependencies
- *
- * foo>=1.0 - 19129 entries
- * foo<1.0 - 1740 entries (only perl)
- * foo>1.0 - 44 entries
- * foo<=2.0 - 1
- * {foo>=1.0,bar>=2.0}
- * foo>=1.0<2.0
- * foo{-bar,-baz}>=1.0
- * foo{-bar,-baz}-[0-9]*
- * foo-{bar,baz}
- * foo-1.0{,nb[0-9]*} - 260
- * foo-[0-9]* - 3214
- * foo-[a-z]*-[0-9]* - not handled here, see pdb_rec_depends()
- * foo-1.0 - 20
- */
-char *
-get_pkgname_from_depend(char *depend)
-{
-	char	*pkgname, *tmp;
-
-	if (depend == NULL || *depend == '\0')
-		return NULL;
-
-	/* 1. worse case, {foo>=1.0,bar-[0-9]*} */
-	if (*depend == '{') {
-		pkgname = xstrdup(depend + 1);
-		tmp = strrchr(pkgname, '}');
-		if (tmp != NULL)
-			*tmp = '\0'; /* pkgname == "foo,bar" */
-
-		/* {foo,bar} should always have comma */
-		while ((tmp = strchr(pkgname, ',')) != NULL)
-			*tmp = '\0'; /* pkgname == foo-[0-9]* or whatever */
-	} else /* we should now have a "normal" pattern */
-		pkgname = xstrdup(depend);
-
-	/* 2. classic case, foo-[<>{?*\[] */
-	clear_pattern(pkgname);
-
-	/* 3. only foo-1.0 should remain */
-	cleanup_version(pkgname);
-
-	return pkgname;
-}
-
 /*
  * Compare string values between two packages.  Often in pkgin the comparison
  * in question allows both to be NULL, for example with pkg_summary fields
@@ -313,3 +270,77 @@ pkgstrcmp(const char *p1, const char *p2)
 
 	return (strcmp(p1, p2));
 }
+
+/*
+ * To speed up lookups of patterns we use hashes keyed on the package name if
+ * available.  This function extracts a single package name from a pattern if
+ * that is the only possible package that can satisfy the pattern.
+ *
+ * For example, "foo-[0-9]*" and "foo>1" etc can be guaranteed to only match
+ * for the package name "foo", whereas "{foo,bar}-[0-9]*", "fo{o,b}*>1" can
+ * not.  The latter return NULL and callers will instead have to traverse the
+ * entire hash.
+ *
+ * This is kept relatively conservative as we cannot be sure what creative
+ * patterns users may come up with in the future.
+ */
+
+char *
+pkgname_from_pattern(const char *pattern)
+{
+	char *p, *pkgname;
+
+	/*
+	 * Any alternate matches can be immediately discounted.  It may
+	 * be possible in the future to expand these and record all
+	 * possible package names to look up in turn.
+	 */
+	if (strpbrk(pattern, "{"))
+		return NULL;
+
+	pkgname = xstrdup(pattern);
+
+	/*
+	 * Since we discounted alternate matches, any specific version
+	 * matches ("foo>1", "foo<5", etc) are guaranteed to have the
+	 * package name on the left of the first match.
+	 *
+	 * The only thing we need to double check is that the package
+	 * name does not contain any globs (not currently used in
+	 * pkgsrc and highly suspicious, but we do not want to take any
+	 * chances).
+	 */
+	if ((p = strpbrk(pkgname, "<>"))) {
+		*p = '\0';
+		if (strpbrk(pkgname, "[]*"))
+			return NULL;
+		return pkgname;
+	}
+
+	/*
+	 * The only other case we support for now is the incredibly common
+	 * "foo-[0-9]*".  Any other version match e.g. "foo-1.*" is a little
+	 * too complicated to handle, just in case there is a "*" glob
+	 * somewhere in the package name, or if the package name itself
+	 * contains '-[0-9]*'.
+	 */
+	if ((p = strrchr(pkgname, '[')) && --p > pkgname) {
+		if (strcmp(p, "-[0-9]*") != 0)
+			return NULL;
+		*p = '\0';
+		if (strpbrk(pkgname, "[]*"))
+			return NULL;
+		return pkgname;
+	}
+
+	return NULL;
+}
+
+/*
+ * qsort callback to sort alphabetically.
+ */
+int
+sort_pkg_alpha(const void *a, const void *b)
+{
+	return strcmp(*(const char * const *)a, *(const char * const *)b);
+}
